use ratatui::prelude::Color;

use crate::dungeon_slayer::MonsterStats;
use crate::position::{HasPosition, Position};

pub trait Monster: HasPosition {
    fn get_name(&self) -> &str;
    fn is_dead(&self) -> bool;
    fn get_representation(&self) -> (&str, Color);
    fn decrease_life(&mut self, by: usize);
    // fn get_immutable_position(&self) -> &Position;
    fn get_experience_gain(&self) -> usize;
    fn get_ticks_between_steps(&self) -> u128;
    #[cfg(test)]
    fn get_life(&self) -> usize;
}

macro_rules! default_monster {
    ($($t:ty),+ $(,)?) => ($(
        impl Monster for $t {
            fn get_name(&self) -> &str { &self.name }
            fn is_dead(&self) -> bool { self.life <= 0 }
            fn get_experience_gain(&self) -> usize { self.experience_gain }
            fn get_representation(&self) -> (&str, Color) { (&self.symbol, self.color) }
            fn decrease_life(&mut self, by: usize) {
                self.life = self.life.saturating_sub(by);
            }
            fn get_ticks_between_steps(&self) -> u128 { self.ticks_between_steps }

            #[cfg(test)]
            fn get_life(&self) -> usize { self.life }
        }
        impl HasPosition for $t {
            fn get_position(&mut self) -> &mut Position {
                &mut self.position
            }
            fn get_immutable_position(&self) -> &Position {
                &self.position
            }
        }
    )+)
}

pub struct Rat {
    name: String,
    life: usize,
    position: Position,
    symbol: String,
    color: Color,
    experience_gain: usize,
    ticks_between_steps: u128,
    monster_stats: MonsterStats,
}

impl Rat {
    pub fn new_with_position(position: Position) -> Self {
        let monster_stats = MonsterStats {
            body: 2,
            agility: 4,
            mind: 1,
            strength: 1,
            toughness: 0,
            movement: 2,
            dexterity: 0,
            wisdom: 0,
            aura: 0,
            max_life: 3,
            defense: 2,
            initiative: 6,
            walk: 3,
            hit: 4,
            shoot: 0,
            cast: 0,
            targeted_cast: 0,
        };
        Self {
            name: "rat".to_string(),
            life: 3,
            position,
            symbol: String::from("R"),
            color: Color::Black,
            experience_gain: 5,
            ticks_between_steps: 5,
            monster_stats,
        }
    }
}
default_monster!(Rat);

pub struct Orc {
    name: String,
    life: usize,
    position: Position,
    symbol: String,
    color: Color,
    experience_gain: usize,
    ticks_between_steps: u128,
    monster_stats: MonsterStats,
}

impl Orc {
    pub fn new_with_position(position: Position) -> Self {
        let monster_stats = MonsterStats {
            body: 10,
            agility: 6,
            mind: 2,
            strength: 2,
            toughness: 3,
            movement: 0,
            dexterity: 3,
            wisdom: 1,
            aura: 0,
            max_life: 23,
            defense: 14,
            initiative: 1,
            walk: 4,
            hit: 13,
            shoot: 10,
            cast: 0,
            targeted_cast: 0,
        };
        Self {
            name: "orc".to_string(),
            life: 23,
            position,
            symbol: String::from("O"),
            color: Color::DarkGray,
            experience_gain: 10,
            ticks_between_steps: 10,
            monster_stats
        }
    }
}

default_monster!(Orc);


#[test]
fn monsters_can_move() {
    let mut m = Rat::new_with_position(Position::new(0, 0, 0));
    assert_eq!(m.get_position(), &Position::new(0, 0, 0));
    m.get_position().change(1, 2);
    assert_eq!(m.get_position(), &Position::new(0, 1, 2));
    m.get_position().change(2, 1);
    assert_eq!(m.get_position(), &Position::new(0, 3, 3));
    m.get_position().set(1, 2, 3);
    m.get_position().change(2, 1);
    assert_eq!(m.get_position(), &Position::new(1, 4, 4));
}

#[test]
fn monsters_can_die() {
    let mut m = Rat::new_with_position(Position::new(0, 0, 0));
    assert_eq!(m.get_life(), 3);
    assert_eq!(m.is_dead(), false);
    m.decrease_life(1);
    assert_eq!(m.get_life(), 2);
    m.decrease_life(3);
    assert_eq!(m.get_life(), 0);
    assert_eq!(m.is_dead(), true);
}

#[test]
fn test_rat_values() {
    let mut m = Rat::new_with_position(Position::new(0, 0, 0));
    assert_eq!(m.monster_stats.max_life, 3);
}